
// Series RLC Circuit Transient Analysis
// =====================================
// Show step response of series RLC circuit
// Example for courses: IEL, IMS
// Method: Runge-Kutta 4
// Warning: Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2017-2020

//
//      o  Uin = 1V
//      |
//      _
//     | |
//     | |  R
//     | |
//      T
//      )
//      )   L
//      )
//      |
//     ===  C
//      |
//      |
//      o  0
//

#include <stdio.h>
#include <stdlib.h>     // strtod()

// System: RLC series
#define SIZE 2                  // state vector size of the system
      double R = 10;            // 10 ohm
const double C = 100e-6;        // 100 uF
const double L = 100e-3;        // 100 mH
const double uin = 1.0;         // step 0->1V at time 0

// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    double u_C = y[0];
    double i   = y[1];
    // Equations transformed to assignments:
    double u_R = R * i;           // Ohm's law
    double u_L = uin - u_C - u_R; // Kirchhoff's voltage law
    yin[0] = i/C;                 // der(u_C)
    yin[1] = u_L/L;               // der(i)
}

// Runge-Kutta-4 step
void RK4step(double t, double h, double *y);

// Experiment control
int main(int argc, char *argv[])
{
    const double TEND = 0.096;
    double y[SIZE] = { 0, 0 };               // initial state
    const double stepsize = 0.00001;
    const unsigned NS = (TEND/stepsize)/500; // print interval
    if(argc>1) {
        R = strtod(argv[1],NULL);
        double RR = R;
        if(R<0)    R=0;
        if(R>200)  R=200;
        if(RR!=R) fprintf(stderr, "Warning: R limited to %g Ohm\n", R);
    }
    printf("# R=%g L=%g C=%g\n", R, L, C);

    unsigned steps = 0;
    double t = 0;
    while (t < TEND) {
        double u_C = y[0];
        double i = y[1];
        double u_R = R * i;
        double u_L = uin - u_C - u_R;
        if(steps%NS == 0)               // output each NS steps
            printf("%8.6f % -11g % -11g  % -11g\n", t, u_C, u_L, i);
        RK4step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
        ++steps;
    }
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}
